'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

function _interopNamespace(e) {
  if (e && e.__esModule) { return e; } else {
    var n = {};
    if (e) {
      Object.keys(e).forEach(function (k) {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () {
            return e[k];
          }
        });
      });
    }
    n['default'] = e;
    return n;
  }
}

const path = _interopDefault(require('path'));
const onExit = _interopDefault(require('signal-exit'));
const kill = _interopDefault(require('tree-kill'));
const vueTemplateCompiler = require('vue-template-compiler');
const crypto = require('crypto');
const fs = _interopDefault(require('fs'));
const util = _interopDefault(require('util'));
const Glob = _interopDefault(require('glob'));
const webpack = _interopDefault(require('webpack'));
const core = require('@babel/core');
const parser = require('@babel/parser');

// https://www.browserstack.com/automate/node#configure-capabilities
// https://wiki.saucelabs.com/display/DOCS/Platform+Configurator

const browsers = [
  'chrome',
  'firefox',
  'ie',
  'jsdom',
  'edge',
  'safari'
];

const browserOptions = [
  'headless',
  'xvfb',
  'staticserver'
];

const drivers = [
  'puppeteer',
  'puppeteer-core',
  'selenium'
];

const providers = [
  'browserstack',
  'saucelabs'
];

const browserVariants = {
  local: ['browserstack'],
  core: ['puppeteer']
};

const os = [
  'windows',
  'linux',
  'mac',
  'macos',
  'macosx',
  'osx',
  'android',
  'ios'
];

function getBrowserConfigFromString(browserString) {
  const browserConfig = {};

  browserString
    .split('/')
    .filter(s => !!s)
    .map(s => s.toLowerCase()
      // replace first occurence with =
      // -> Chrome 39 => chrome=39
      // -> Mac OS X=High Sierra => Mac OS X=High Sierra
      .replace(/([\s:])/, (m, $1, o, s) => (s.includes('=') ? $1 : '='))
    )
    .forEach((s) => {
      let [key, value] = s.split('=', 2); // eslint-disable-line prefer-const
      key = key.replace(/\s+/, '');

      if (browsers.includes(key)) {
        browserConfig.browser = key;

        if (value) {
          browserConfig.browserVersion = value.replace(/\s+/g, '');
        }
        return
      }

      if (browserOptions.includes(key)) {
        browserConfig[key] = typeof value === 'undefined' ? true : value;
        return
      }

      if (os.includes(key)) {
        browserConfig.os = key;

        if (value) {
          browserConfig.osVersion = value;
        }
        return
      }

      if (drivers.includes(key)) {
        browserConfig.driver = key;
        return
      }

      if (providers.includes(key)) {
        browserConfig.provider = key;
        return
      }

      if (browserVariants[key]) {
        browserConfig.browserVariant = key;
        return
      }

      if (key === 'device') {
        browserConfig.device = value;
        return
      }

      if (!value) {
        const screenResolution = key.split(/(\d+)x(\d+)/i);
        if (screenResolution && screenResolution.length > 1) {
          browserConfig.window = {
            width: parseInt(screenResolution[1]),
            height: parseInt(screenResolution[2])
          };
        }
      }
    });

  return browserConfig
}

function getBrowserImportFromConfig(browserConfig) {
  const importPath = [];

  if (browserConfig.browser) {
    importPath.push(browserConfig.browser);
  }

  if (browserConfig.driver && !browserConfig.provider) {
    importPath.push(browserConfig.driver);
  }

  if (browserConfig.provider) {
    importPath.push(browserConfig.provider);
  }

  if (browserConfig.browserVariant) {
    importPath.push(browserConfig.browserVariant);
  }

  return importPath.join('/')
}

function getDefaultHtmlCompiler() {
  return html => vueTemplateCompiler.compile(html).ast
}

const md5sum = str => crypto.createHash('md5').update(str).digest('hex');

const readDir = util.promisify(fs.readdir);
const readFile = util.promisify(fs.readFile);
const glob = util.promisify(Glob);

function exists(p) {
  return new Promise((resolve, reject) => {
    fs.access(p, fs.constants.F_OK, (err) => {
      if (err) {
        resolve(false);
        return
      }

      resolve(true);
    });
  })
}

function stats(p) {
  return new Promise((resolve, reject) => {
    fs.stat(p, (err, stats) => {
      if (err) {
        resolve(false);
        return
      }

      resolve(stats);
    });
  })
}

function requireResolve(path) {
  /* istanbul ignore next */
  return require.resolve(path)
}

let nodeModulesPath;

async function checkNodeModulesPath(modulesPath) {
  const pathExists = await exists(modulesPath);

  if (pathExists) {
    nodeModulesPath = modulesPath;
    return true
  }

  return false
}

async function findNodeModulesPath(refresh) {
  if (nodeModulesPath && !refresh) {
    return nodeModulesPath
  }

  let modulesPath = path.resolve(__dirname, '../../..', 'node_modules');
  if (await checkNodeModulesPath(modulesPath)) {
    return nodeModulesPath
  }

  modulesPath = path.resolve('./node_modules');
  if (await checkNodeModulesPath(modulesPath)) {
    return nodeModulesPath
  }

  const hablePath = requireResolve('hable/package.json');
  modulesPath = path.dirname(path.dirname(hablePath));
  if (await checkNodeModulesPath(modulesPath)) {
    return nodeModulesPath
  }

  modulesPath = __dirname;
  while (modulesPath.length > 1) {
    const tryPath = path.join(modulesPath, 'node_modules');
    modulesPath = path.dirname(modulesPath);

    if (await checkNodeModulesPath(tryPath)) {
      modulesPath = tryPath;
      break
    }
  }

  if (modulesPath.length <= 1) {
    modulesPath = '';
  }

  nodeModulesPath = modulesPath;
  return modulesPath
}

function createCacheKey(fn, opts) {
  if (typeof opts !== 'string') {
    opts = JSON.stringify(opts);
  }

  return md5sum(`${fn}-x-${opts}`)
}

async function getCachePath(filePath = '') {
  const modulesPath = await findNodeModulesPath();
  return path.join(modulesPath, '.cache', 'tib', filePath)
}

class BrowserError extends Error {
  constructor(classInstance, message, ...params) {
    let className;
    if (!message) {
      message = classInstance;
      className = 'BrowserError';
    } else if (typeof classInstance === 'object') {
      className = classInstance.constructor.name;
    } else if (typeof classInstance === 'string') {
      className = classInstance;
    }

    // Pass remaining arguments (including vendor specific ones) to parent constructor
    super(`${className}: ${message}`, ...params);

    // Maintains proper stack trace for where our error was thrown (only available on V8)
    if (Error.captureStackTrace) {
      Error.captureStackTrace(this, BrowserError);
    }
  }
}

const camelCase = str => str.replace(/(^|[-_.\s]+)(.)/g, (m, $1, $2, offset) => !offset ? $2.toLowerCase() : $2.toUpperCase());

function abstractGuard(className, { name } = {}) {
  if (className === name) {
    throw new BrowserError(`Do not use abstract class '${className}' directly`)
  }
}

function isMockedFunction(fn, fnName) {
  return fn.name !== fnName
}

async function loadDependency(dependency) {
  try {
    const module = await new Promise(function (resolve) { resolve(_interopNamespace(require(dependency))); }).then(m => m.default || m);
    return module
  } catch (e) {
    throw new BrowserError(`Could not import the required dependency '${dependency}'
(error: ${e.message})

Please install the dependency with:
$ npm install ${dependency}
or
$ yarn add ${dependency}
    `)
  }
}

async function createPageFunctions(page, sourceFiles, babelPresets) {
  const pageFunctions = {};

  if (!sourceFiles) {
    return pageFunctions
  }

  if (typeof sourceFiles === 'string') {
    sourceFiles = await glob(sourceFiles);
  }

  for (let file of sourceFiles) {
    let fnName;
    if (Array.isArray(file)) {
      [file, fnName] = file;
    } else {
      fnName = camelCase(path.basename(file, '.js'));
    }

    if (pageFunctions[fnName]) {
      // eslint-disable-next-line no-console
      console.warn(`A page function with name '${fnName}' already exists, the previous one will be overwritten`);
    }

    pageFunctions[fnName] = async (...args) => {
      const body = await getPageFunctionBody(fnName, file, babelPresets);

      return page.runAsyncScript({
        args: [],
        body: `${body} return (pageFn.${fnName} || pageFn.default).apply(null, arguments)`
      }, ...args)
    };
  }

  return pageFunctions
}

async function getPageFunctionBody(fnName, filePath, babelPresets) {
  const cacheKey = createCacheKey(filePath, babelPresets);
  const cacheFile = `${fnName}-${cacheKey}.js`;
  const cachePath = await getCachePath(cacheFile);
  let cacheValid = await exists(cachePath);

  if (cacheValid) {
    const { mtime: fileModified } = await stats(filePath);
    const { mtime: cacheModified } = await stats(cachePath);

    cacheValid = cacheModified > fileModified;
  }

  if (!cacheValid) {
    // transpile page function
    await compilePageFunction({
      name: 'pageFn',
      filePath,
      cachePath,
      babelPresets
    });
  }

  const fnBody = await readFile(cachePath, { encoding: 'utf8' });
  return fnBody
}

async function compilePageFunction({ babelPresets, ...config }) {
  let babelOptions;

  if (babelPresets) {
    babelOptions = {
      presets: [
        ['@babel/preset-env', babelPresets]
      ]
    };
  } else {
    babelOptions = {
      babelrcRoots: [
        '.',
        path.resolve(await findNodeModulesPath(), '..')
      ]
    };
  }

  config.babelOptions = babelOptions;

  return new Promise((resolve, reject) => {
    const webpackConfig = createWebpackConfig(config);

    webpack(webpackConfig, (err, stats) => {
      /* istanbul ignore next */
      if (err) {
        reject(err);
        return
      }

      if (stats.hasErrors()) {
        const [err] = stats.compilation.errors;
        reject(err);
        return
      }

      resolve(stats);
    });
  })
}
function createWebpackConfig({
  name,
  filePath,
  cachePath,
  babelOptions = {}
}) {
  return {
    mode: 'production',
    entry: filePath,
    output: {
      path: path.dirname(cachePath),
      filename: path.basename(cachePath),
      library: name,
      libraryTarget: 'var'
    },
    plugins: [
      new webpack.optimize.LimitChunkCountPlugin({
        maxChunks: 1
      })
    ],
    module: {
      rules: [
        {
          test: /\.js$/,
          exclude: /node_modules/,
          use: {
            loader: 'babel-loader',
            options: babelOptions
          }
        }
      ]
    }
  }
}

const fnCache = {};

function parseFunction(fn, args, presetOptions) {
  if (arguments.length === 2) {
    presetOptions = args || {};
    args = undefined;
  }

  if (typeof fn !== 'function') {
    throw new BrowserError(`parseFunction expects the first argument to be a function, received '${typeof fn}' instead`)
  }

  const fnString = fn.toString();
  const cacheKey = createCacheKey(fnString, presetOptions);

  if (fnCache[cacheKey] && fnCache[cacheKey]) {
    return fnCache[cacheKey]
  }

  const parsed = {};
  let ast = parser.parseExpression(fnString);
  parsed.args = ast.params.map(p => p.name);

  ast = ast.body;

  /* transform can only transform a program and a program is an Array of
   * Statements.
   * The body of an arrow function like 'arg => arg' does not
   * contain a Statement, so we add a block & return statement in that case.
   * The return statement is needed later when we create a new Function and the
   * block statement helps so we can (almost) always call slice(1, -1) below to
   * retrieve the real function body
   */
  if (!ast.type.includes('Statement')) {
    ast = core.types.blockStatement([core.types.returnStatement(ast)]);
  }

  const transpiled = core.transformFromAstSync(
    core.types.file(core.types.program([ast])),
    undefined,
    {
      sourceType: 'script',
      presets: [
        ['@babel/preset-env', presetOptions]
      ]
    }
  );

  if (transpiled.code[0] === '{') {
    // remove block statement needed to transform & trim block whitespace
    parsed.body = transpiled.code.slice(1, -1).trim();
  } else {
    /* istanbul ignore next */
    parsed.body = transpiled.code;
  }

  fnCache[cacheKey] = parsed;
  return parsed
}

// TODO: add more test framework checks like sinon?
function disableTimers() {
  // find Jest fingerprint
  if (process.env.JEST_WORKER_ID) {
    try {
      jest.useFakeTimers();
    } catch (e) {
      /* istanbul ignore next */
      throw new BrowserError(`Enabling fake timers failed: ${e.message}`)
    }
  }
}

function enableTimers() {
  // find Jest fingerprint
  if (process.env.JEST_WORKER_ID) {
    try {
      // call useFakeTimers first as it seems there is no way to determine
      // if fake timers are used at all and otherwise runOnlyPendingTimers
      // will fail with a warning from within Jest
      // -> Disabled because we probably dont care about pending timers
      // jest.useFakeTimers()
      // jest.runOnlyPendingTimers()
      jest.useRealTimers();
    } catch (e) {
      /* istanbul ignore next */
      throw new BrowserError(`Enabling real timers failed: ${e.message}`)
    }
  }
}

const consola = console; // eslint-disable-line no-console

let PID;

class BrowserStackLocal {
  static async loadDriver() {
    if (BrowserStackLocal.driver) {
      return
    }

    const browserstack = await loadDependency('browserstack-local');
    BrowserStackLocal.driver = new browserstack.Local();
  }

  static async start(config = {}) {
    // TODO: support webserver
    if (!config.folder && config.folder !== false) {
      config.folder = path.resolve(process.cwd());
    }

    await BrowserStackLocal.loadDriver();

    // util.promisify doesnt work due to this binding and
    // .bind(local) didnt work either
    return new Promise((resolve, reject) => {
      BrowserStackLocal.driver.start(config, (error) => {
        if (error) {
          reject(error);
        }

        onExit(() => BrowserStackLocal.stop());

        PID = BrowserStackLocal.driver.pid;
        resolve(PID);
      });
    })
  }

  static stop(pid) {
    pid = pid || (BrowserStackLocal.driver && BrowserStackLocal.driver.pid) || PID;

    if (!BrowserStackLocal.driver || !pid) {
      consola.warn('Stop called but browserstack-local was not started');
      return
    }

    return new Promise((resolve, reject) => {
      // local.stop is buggy, it doesnt kill anything and takes forever
      // after looking at the local.stop implementation tree-kill does
      // practically the same
      kill(pid, 'SIGTERM', (error) => {
        if (error) {
          /* istanbul ignore next */
          reject(error);
        }

        resolve();
      });
    })
  }
}

exports.BrowserError = BrowserError;
exports.BrowserStackLocal = BrowserStackLocal;
exports.abstractGuard = abstractGuard;
exports.createPageFunctions = createPageFunctions;
exports.disableTimers = disableTimers;
exports.enableTimers = enableTimers;
exports.getBrowserConfigFromString = getBrowserConfigFromString;
exports.getBrowserImportFromConfig = getBrowserImportFromConfig;
exports.getDefaultHtmlCompiler = getDefaultHtmlCompiler;
exports.isMockedFunction = isMockedFunction;
exports.loadDependency = loadDependency;
exports.parseFunction = parseFunction;
